""" Tests of PackageTester itself. """

# pylint: disable=E0401, R0915, W0511


import logging
import re
import sys

import pytest
from package_tester import package_tester
from package_tester.package_tester import PackageTesterException, build_usage_message
from package_tester.package_tester_utils import (
    execute_shell_command,
    ENTRYPOINT,
    PytestResultsParser,
    ResultType,
    running_in_shell,
)

logger = logging.getLogger(__name__)

_MIN_OUTPUT_LINE_COUNT = 7
_RUN_ALL = True


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_entrypoint_finds_valid_package():
    """
    If a package exists, it ought to be found and tested
    in both regular and verbose mode.

    :return:
    """
    # control: we know these will pass
    package_name = "dummy_tests.always_pass"

    output = execute_shell_command(f"{ENTRYPOINT} {package_name}")
    assert len(output) == 1
    summary_line = output[-1]
    summary = PytestResultsParser(summary_line).parse()
    assert summary.passes == 4
    assert summary.skips == summary.errors == summary.failures == 0
    assert summary.elapsed_time_numeric() > 0.0


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_main_finds_valid_package():
    """
    If a package exists, it ought to be found and tested
    in both regular and verbose mode.

    :return:
    """

    # control: we know this one will work
    package_name = "dummy_tests.always_pass"

    # fake up command-line parameters
    sys.argv.clear()
    sys.argv.insert(0, package_name)
    sys.argv.insert(0, __file__)
    tester = package_tester.main()
    assert len(tester.pytested_packages) == 1
    package_info = tester.pytested_packages[0]
    for line in package_info.test_output[:1]:
        assert "error" not in line.lower()
    summary = package_info.summary
    assert summary.passes == 4
    assert summary.skips == summary.errors == summary.failures == 0
    assert summary.elapsed_time_numeric() > 0.0

    # with verbose option
    sys.argv.clear()
    sys.argv.insert(0, "-v")
    sys.argv.insert(0, package_name)
    sys.argv.insert(0, __file__)
    tester = package_tester.main()
    results = tester.pytested_packages
    assert len(results) == 1
    package_info = tester.pytested_packages[0]
    assert len(package_info.test_output) == 12

    # summary should be identical
    verbose_summary1 = package_info.summary
    assert verbose_summary1 == package_info.summary

    # result should be same with '-v' after entrypoint
    sys.argv.clear()
    sys.argv.insert(0, package_name)
    sys.argv.insert(0, "-v")
    sys.argv.insert(0, __file__)
    tester = package_tester.main()
    results = tester.pytested_packages
    assert len(results) == 1
    package_info = tester.pytested_packages[0]
    assert len(package_info.test_output) == 12
    verbose_summary2 = package_info.summary
    assert verbose_summary2 == verbose_summary1

    # now a real package
    package_name = "datafetcher"

    # will succeed in IJ and Docker container but fail at command line
    if running_in_shell():
        return

    sys.argv.clear()
    sys.argv.insert(0, package_name)
    sys.argv.insert(0, __file__)
    tester = package_tester.main()
    results = tester.pytested_packages
    assert len(results) == 1
    package_info = tester.pytested_packages[0]
    summary = package_info.summary
    assert summary.passes >= 0
    assert summary.skips >= 0
    assert summary.errors >= 0
    assert summary.failures >= 0
    assert summary.elapsed_time_numeric() > 0.0
    # be sure at least one test actually got run
    assert summary.passes + summary.skips + summary.errors + summary.failures > 0


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_main_reports_package_not_found():
    """
    Giving the tester a bogus package name should produce a useful error.

    :return:
    """
    package_name = "moe.larry.curley"

    sys.argv.clear()
    sys.argv.insert(0, package_name)
    sys.argv.insert(0, __file__)

    with pytest.raises(PackageTesterException) as exc:
        package_tester.main()
    exc_value = "".join(exc.value.args)
    assert exc_value == "expected 1 packages; got 0"


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_entrypoint_reports_package_not_found():
    """
    Entrypoint with bogus package should output expected failure message

    :return:
    """
    package_name = "moe.larry.curley"
    output = execute_shell_command(f"{ENTRYPOINT} {package_name}")
    assert len(output) > _MIN_OUTPUT_LINE_COUNT
    assert output[0].startswith("Traceback")
    the_end = output[-1]
    assert (
        the_end == "package_tester.package_tester.PackageTesterException: "
        "expected 1 packages; got 0"
    )

    # verbosely; should have same result
    verbose_output = execute_shell_command(f"{ENTRYPOINT} {package_name} -v")
    assert verbose_output[0] == output[0]
    last_dying_breath = verbose_output[-1]
    assert last_dying_breath == the_end


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_segregates_package_results():
    """
    When we give PackageTester multiple packages,
    does it keep results separate?

    :return:
    """
    package_names = ["always_fail", "always_pass"]
    sys.argv.clear()
    for package_name in package_names:
        sys.argv.insert(0, package_name)
    sys.argv.insert(0, __file__)
    tester = package_tester.main()

    packages_info = tester.pytested_packages
    assert len(packages_info) == len(package_names)
    for package_info in packages_info:
        summary = package_info.summary
        assert not summary.is_empty()
        assert len(package_info.test_output) >= 2
        if "fail" in package_info.package_name:
            assert summary.failures == 2
            assert summary.passes == 1
        else:
            assert summary.failures == 0
            assert summary.passes == 4

    output = tester.output
    assert len(output) >= 3
    num_locations_found = 0
    for line in output:
        match = re.search(r"([0-9]+) Package locations found", line)
        if match:
            num_locations_found = int(match.group(1))
            break
    assert num_locations_found == len(package_names)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_entrypoint_launches_tester():
    """
    Can we launch PackageTester at command line using 'package-tester'?

    :return:
    """
    package_name = "test.dummy_tests.always_pass"
    command = f"package-tester {package_name}"
    output = execute_shell_command(command)
    assert len(output) >= 1
    summary_line = None
    passed = None
    for line in output:
        if (
            ResultType.PASSED.value in line or ResultType.PASSES.value in line
        ) and "elapsed time" in line:
            passed = (
                ResultType.PASSED.value
                if ResultType.PASSED.value in line
                else ResultType.PASSES.value
            )
            summary_line = line
            break
    assert summary_line

    assert "4 " + passed in summary_line
    assert "0 errors" in summary_line
    assert "0 skipped" in summary_line


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_main_handles_failure_accurately():
    """
    When any of the tests fails, the tester should store the output and
    results as expected.

    :return:
    """
    package_name = "dummy_tests"
    sys.argv.clear()
    sys.argv.insert(0, package_name)
    sys.argv.insert(0, __file__)
    tester = package_tester.main()
    output = tester.output
    assert output is not None
    assert len(output) > 0

    package_info = tester.pytested_packages[0]
    assert package_info.summary.failures == 2

    failure_count = 0
    for line in package_info.test_output:
        if "failed" in line.lower():
            failure_count += 1
    assert failure_count >= 2


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_entrypoint_handles_failure_accurately():
    """
    Summary should contain correct number of test failures.

    :return:
    """
    package_name = "dummy_tests"

    # terse (default)
    output = execute_shell_command(f"{ENTRYPOINT} {package_name}")
    assert len(output) == 1
    summary = PytestResultsParser(output[-1]).parse()
    assert summary.passes == 5
    assert summary.failures == 2
    # verbose
    verbose_output = execute_shell_command(f"{ENTRYPOINT} -v {package_name}")
    assert len(verbose_output) > _MIN_OUTPUT_LINE_COUNT

    pass_count = failure_count = error_count = skip_count = 0
    for line in verbose_output[:-2]:
        if "=== FAILURES ===" in line:
            # individual test results are complete
            break

        if "PASSED" in line:
            pass_count += 1
        elif "FAILED" in line:
            failure_count += 1
        elif "SKIPPED" in line:
            skip_count += 1
        elif "ERROR" in line:
            error_count += 1

    assert pass_count == 5
    assert failure_count == 2
    assert error_count == skip_count == 0

    verbose_summary = PytestResultsParser(verbose_output[-1]).parse()
    assert verbose_summary == summary


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_main_fails_if_no_args():
    """
    Given any scenario involving no package name, tester should
    say exactly what the problem was.

    :return:
    """

    expected_message = "One or more package names required\n" + build_usage_message()

    # no args whatsoever
    sys.argv.clear()
    with pytest.raises(PackageTesterException) as exc:
        package_tester.main()
    actual_message = "".join(exc.value.args)
    assert actual_message == expected_message

    # __file__ as only arg
    sys.argv.insert(0, __file__)
    with pytest.raises(PackageTesterException) as exc:
        package_tester.main()
    actual_message = "".join(exc.value.args)
    assert actual_message == expected_message

    # package is None
    sys.argv.insert(1, None)
    with pytest.raises(PackageTesterException) as exc:
        package_tester.main()
    actual_message = "".join(exc.value.args)
    assert actual_message == expected_message

    # empty packages
    sys.argv.insert(1, "")
    sys.argv.insert(2, "")
    with pytest.raises(PackageTesterException) as exc:
        package_tester.main()
    actual_message = "".join(exc.value.args)
    assert actual_message == expected_message


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_entrypoint_fails_if_no_args():
    """
    Entrypoint should fail predictably when args are omitted

    :return:
    """
    expected_message = "One or more package names required\n" + build_usage_message()
    expected_output_lines = expected_message.split("\n")
    # no args whatsoever
    output = execute_shell_command(f"{ENTRYPOINT}")
    assert len(output) >= len(expected_output_lines)
    initial_output_lines = output[: len(expected_output_lines) - 1]

    # before comparing, remove any blank lines; they don't count
    initial_output_lines = [line for line in initial_output_lines if line]
    expected_output_lines = [line for line in expected_output_lines if line]

    assert initial_output_lines == expected_output_lines


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_verbose_mode_stores_same_data_as_regular():
    """
    Whether or not the verbose flag is used, all the same output should
    be stored; verbose should print all the output
    rather than just the summary.

    :return:
    """
    package_name = "dummy_tests"
    sys.argv.clear()
    sys.argv.insert(0, package_name)
    sys.argv.insert(0, __file__)
    tester = package_tester.main()
    pytested_packages = tester.pytested_packages
    assert len(pytested_packages) == 1
    regular_info = pytested_packages[0]
    regular_output = regular_info.test_output

    sys.argv.insert(len(sys.argv), "-v")
    tester = package_tester.main()
    verbose_pytested_packages = tester.pytested_packages
    assert len(verbose_pytested_packages) == 1
    verbose_info = verbose_pytested_packages[0]
    assert verbose_info.summary == regular_info.summary

    verbose_output = verbose_info.test_output
    assert len(verbose_output) == len(regular_output)


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_entrypoint_same_output_as_direct_invocation():
    """
    Whether we call package_tester.main() programmatically or launch it using
    the 'package-tester' entrypoint, output should be the same.

    :return:
    """
    package_name = "dummy_tests"

    sys.argv.clear()
    sys.argv.insert(0, package_name)
    sys.argv.insert(0, __file__)
    tester = package_tester.main()
    packages_info = tester.pytested_packages
    assert len(packages_info) == 1

    command = f"{ENTRYPOINT} {package_name}"
    entrypoint_output = execute_shell_command(command)

    entrypoint_summary = PytestResultsParser(entrypoint_output[0]).parse()

    assert entrypoint_summary == packages_info[0].summary


@pytest.mark.skipif(not _RUN_ALL, reason="debugging selected test(s)")
def test_entrypoint_handles_verbose_mode():
    """
    Make sure PackageTester understands verbose mode when invoked from the
    command line using 'package-tester'

    :return:
    """
    package_name = "dummy_tests"
    command = f"package-tester {package_name}"
    regular_output = execute_shell_command(command)

    command = f"package-tester {package_name} -v"
    verbose_output1 = execute_shell_command(command)
    expected_output_len = 33 if running_in_shell() else 32
    assert len(verbose_output1) == expected_output_len
    assert len(regular_output) == 1

    regular_summary = PytestResultsParser(regular_output[-1]).parse()
    verbose_summary1 = PytestResultsParser(verbose_output1[-1]).parse()
    assert verbose_summary1 == regular_summary

    # result should be same if flag is in the middle rather than at end
    command = f"package-tester -v {package_name}"
    verbose_output2 = execute_shell_command(command)
    verbose_summary2 = PytestResultsParser(verbose_output2[-1]).parse()
    assert verbose_summary2 == regular_summary
    assert len(verbose_output2) == expected_output_len
